# This script contain functions help quantization
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing

import json

from tqdm import tqdm
from .operators import *
from .utils import *

# global var
quant_layer_list = []


def quant_ops_switch(name, module, bit_width, logger):
    """quantization operators switch func

    Args:
        name (str): module name
        module (nn.module): module object
        bit_width (int): quantize bit width
        logger (logger): logger for logging

    Returns:
        QuantClass: specific QuantClass for specific layer type
    """
    if isinstance(module, TorchFuncTuple):
        quant_layer = QuantTorchFunc(name, module, bit_width, logger)
    elif isinstance(module, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
        quant_layer = QuantConv(name, module, bit_width, logger)
    elif isinstance(module, nn.Linear):
        quant_layer = QuantLinear(name, module, bit_width, logger)
    elif isinstance(module, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)):
        quant_layer = QuantBatchNorm(name, module, bit_width, logger)
    elif isinstance(module, (nn.LayerNorm)):
        quant_layer = QuantLayerNorm(name, module, bit_width, logger)
    elif isinstance(module, nn.ReLU):
        quant_layer = QuantRelu(name, module, bit_width, logger)
    elif isinstance(
            module,
        (nn.AdaptiveAvgPool1d, nn.AdaptiveAvgPool2d, nn.AdaptiveAvgPool3d)):
        quant_layer = QuantAdaptiveAvgPool(name, module, bit_width, logger)
    else:
        quant_layer = None

    return quant_layer


def quantize(model,
             bit_width,
             data_loader,
             path,
             fuse_norm=True,
             is_debug=False,
             to_onnx=False):
    """quantize main func

    Args:
        model (nn.module): Pytorch model
        bit_width (int): Quantize bit width
        data_loader (Dataloader): model data_loader
        path (str): log path for saving quant params, quantized model, log file, etc.
        fuse_norm (bool, optional): if True, fuse normalization module in model. Defaults to True.
        is_debug (bool, optional): if True, turn on debug mode and generate figure of layer stats. Defaults to False.
    """
    # init abstract log path and logger
    logger, log_path = lognow(path)
    # fuse normalization layer
    if fuse_norm:
        fuse_module(model)
    # if debug
    if is_debug:
        logger.warning("Debug Mode:")
        logger.warning("Init quant_layer_list.")
        logger.warning("Draw histograms figure of input.")
        quant_input = QuantInput("input", None, bit_width, logger, log_path,
                                 data_loader)
        quant_input.debug_input()
        logger.warning("Draw histograms figure of weight.")
        logger.warning("Add get_activation hook.")
        for name, module in model.named_modules():
            # if bool(list(module.children())) is False, then module can't be container
            if not list(module.children()):
                quant_layer = QuantDebug(name, module, bit_width, logger,
                                         log_path)
                quant_layer_list.append(quant_layer)
                quant_layer.quantize_weight()
                quant_layer.register_get_activation_hook()

        logger.warning("Init boundary.")
        for i, (image, index) in enumerate(tqdm(data_loader)):
            image = image.cuda()
            model(image)
            for layer in quant_layer_list:
                layer.init_boundary()

        logger.warning("Init histograms of activation.")
        for i, (image, index) in enumerate(tqdm(data_loader)):
            image = image.cuda()
            model(image)
            for layer in quant_layer_list:
                layer.init_histograms()

        logger.warning("Draw histograms figure of activation.")
        for layer in quant_layer_list:
            layer.quantize_activation()
    else:
        logger.info("Init quant_layer_list.")
        logger.info("Quantize the input.")
        quant_input = QuantInput("input", None, bit_width, logger, log_path,
                                 data_loader)
        quant_input.quantize_input()
        logger.info("Add get_activation hook.")
        logger.info("Quantize the weight.")
        for name, module in model.named_modules():
            quant_layer = quant_ops_switch(name, module, bit_width, logger)
            if quant_layer:
                quant_layer_list.append(quant_layer)
                quant_layer.quantize_weight()
                quant_layer.register_get_activation_hook()
        # save quantized weight
        torch.save(model.state_dict(), log_path + '/quantized.pth')

        logger.info("Collect histograms of activations.")
        for i, (image, index) in enumerate(tqdm(data_loader)):
            image = image.cuda()
            model(image)
            for layer in quant_layer_list:
                layer.init_histograms()

        logger.info("Quantize the activation.")
        for layer in quant_layer_list:
            layer.quantize_activation()

        logger.info("Turn on eval mode.")
        for layer in quant_layer_list:
            layer.eval_mode()

        # insert quant_input to quant_layer_list
        quant_layer_list.insert(0, quant_input)

        # save quant_params
        fl_dict = {}
        for layer in quant_layer_list:
            fl_dict[layer.name] = {}
            fl_dict[layer.name]['weight_fl'] = int(layer.weight_fl)
            fl_dict[layer.name]['activation_fl'] = int(layer.activation_fl)
        with open(log_path + '/quant_params.json', 'w+') as file:
            json.dump(fl_dict, file)


def quantize_parallel(net, bit_width, data_loader, logger, log_path):
    logger.info("Init quant_layer_list.")
    logger.info("Add get_activation hook.")
    logger.info("Quantize the weight.")
    for name, module in net.named_modules():
        quant_layer = op_switch(name, module, bit_width)
        if quant_layer:
            quant_layer_list.append(quant_layer)
    pool = Pool()
    pool.map(lambda x: x.quantize_weight(), quant_layer_list)
    pool.map(lambda x: x.register_get_activation_hook(), quant_layer_list)
    pool.close()
    pool.join()
    torch.save(net.state_dict(), log_path + '/quantized.pth')

    logger.debug("Debug Mode: init boundary")
    if True:
        for i, (image, index) in enumerate(tqdm(data_loader)):
            image = image.cuda()
            net(image)
            pool = Pool()
            pool.map(lambda x: x.init_boundary(), quant_layer_list)
            pool.close()
            pool.join()

    logger.info("Collect histograms of activations.")
    for i, (image, index) in enumerate(tqdm(data_loader)):
        image = image.cuda()
        net(image)
        pool = Pool(12)
        pool.map(lambda x: x.init_histograms(), quant_layer_list)
        pool.close()
        pool.join()

    logger.info("Quantize the activation.")
    pool = Pool()
    pool.map(lambda x: x.quantize_activation(), quant_layer_list)
    pool.close()
    pool.join()



from .activation import *
from .convolution import *
from .dropout import *
from .linear import *
from .normalization import *
from .padding import *
from .pooling import *
from .recurrent import *
from .sparse import *
from .torchfunc import *
from .transformer import *
from .vision import *


# This script contain util will help you, good luck!
# Author:
#     Albert Dongz
# History:
#     2020.4.17 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing

import os
import torch
import logging
import datetime
import torch.nn as nn


def lognow(path="./checkpoints/"):
    """This function set up logging.

    Keyword Arguments:
        path {str} -- log save path (default: {"./checkpoints/"})

    Returns:
        str -- log absolute path, contain time and param path.
    """
    # get current time and create log dir
    now = datetime.datetime.now().strftime('%Y-%m-%d--%H-%M-%S')
    log_path = os.path.normpath(path + '/' +now)
    if not os.path.exists(log_path):
        os.makedirs(log_path)

    # setting up logging
    logger = logging.getLogger(__name__)
    logger.setLevel(level=logging.INFO)
    handler = logging.FileHandler(log_path + "/log.txt")
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    logger.addHandler(handler)
    logger.addHandler(console)
    logger.info('Start logging.')
    return logger, log_path


def float2fixed(data, bw, fl):
    """turn torch.Tensor to fixed Tensor.
    #TODO add stochastic rounding.

    Args:
        data (torch.Tensor): torch.Tensor
        bw (int): bit width of quantization
        fl (int): fraction length of data, from left to right, ignore sign bit.

    Returns:
        torch.Tensor: fixed Tensor
    """
    unit = 2.0**(-bw + 1 + fl)
    fixed_data = torch.clamp(torch.round(data / unit) * unit,
                             min=-2**fl + unit,
                             max=+2**fl - unit)
    return fixed_data


def bn2scale(model_state, layer_name):
    """
    description:
        将指定bn层的函数改成scale层的形式，解决由PyTorch BN层的四个参数数量级不一致导致的在量化时的精度
        较大损失，返回值为包含处理好的bn层的模型参数。
        PyTorch BN数学公式：out = gamma * (x - mean)/sqrt(var) + beta
        Caffe Scale 数学公式：out = alpha * x + beta
        通过使得 new_gamma = gamma / sqrt(var),
               new_beta = beta - gamma * mean / sqrt(var)
               new_mean = 0, new_var = 1 来完成变换。P.S. 为了利用原来的BN的结构来实现scale。
    :parameter
        model_state:
            未处理过的模型参数
        layer_name: str
            bn层的参数存储名字的layername，如cnn.batchnorm0，如果全名为cnn.batchnorm0.weight
    :return:
        model_state:
            处理过的模型参数
    """
    # 从未处理的模型中获取bn层的四个参数
    weight = model_state.get(layer_name + ".weight")
    bias = model_state.get(layer_name + ".bias")
    mean = model_state.get(layer_name + ".running_mean")
    var = model_state.get(layer_name + ".running_var")
    # 计算标准差，即sqrt(var) => std
    std = torch.sqrt(var)
    # 计算新的值
    a = weight / std
    b = bias - weight * mean / std
    e = torch.zeros(mean.shape).to(torch.int32)
    s = torch.ones(var.shape).to(torch.int32)
    # 保存参数到模型
    model_state[layer_name + ".weight"] = a
    model_state[layer_name + ".bias"] = b
    model_state[layer_name + ".running_mean"] = e
    model_state[layer_name + ".running_var"] = s


def bn2scale(module):
    """
    description:
        将指定bn层的函数改成scale层的形式，解决由PyTorch BN层的四个参数数量级不一致导致的在量化时的精度
        较大损失，返回值为包含处理好的bn层的模型参数。
        PyTorch BN数学公式：out = gamma * (x - mean)/sqrt(var) + beta
        Caffe Scale 数学公式：out = alpha * x + beta
        通过使得 new_gamma = gamma / sqrt(var),
               new_beta = beta - gamma * mean / sqrt(var)
               new_mean = 0, new_var = 1 来完成变换。P.S. 为了利用原来的BN的结构来实现scale。
    :parameter
        model_state:
            未处理过的模型参数
        layer_name: str
            bn层的参数存储名字的layername，如cnn.batchnorm0，如果全名为cnn.batchnorm0.weight
    :return:
        model_state:
            处理过的模型参数
    """
    # 从未处理的module中获取bn层的四个参数
    weight = module.weight
    bias = module.bias
    mean = module.running_mean
    var = module.running_var
    # 计算标准差，即sqrt(var) => std
    std = torch.sqrt(var)
    # 计算新的值
    new_weight = weight / std
    new_bias = bias - weight * mean / std
    new_mean = torch.zeros(mean.shape)
    new_var = torch.ones(var.shape)
    # 保存参数到模型
    module.weight = nn.Parameter(new_weight)
    module.bias = nn.Parameter(new_bias)
    module.running_mean = new_mean
    module.running_var = new_var


def transfer_bn2scale(net):
    """Auto transfer BatchNorm to Scale

    Arguments:
        net {torch.nn.Module} -- net need be processed
        state {OrderDict} -- checkpoint of net

    Returns:
        net, state -- processed net and state
    """
    print("Start transfer BatchNorm to Scale")
    for name, module in net.named_modules():
        if isinstance(module, nn.BatchNorm2d):
            bn2scale(module)
    print("Transfer done.")


# Ref: https://zhuanlan.zhihu.com/p/49329030
# fuse BN and Conv
def fuse_module(m):
    class DummyModule(nn.Module):
        def __init__(self):
            super(DummyModule, self).__init__()

        def forward(self, x):
            return x

    def fuse(conv, bn):
        w = conv.weight
        mean = bn.running_mean
        var_sqrt = torch.sqrt(bn.running_var + bn.eps)

        beta = bn.weight
        gamma = bn.bias

        if conv.bias is not None:
            b = conv.bias
        else:
            b = mean.new_zeros(mean.shape)

        w = w * (beta / var_sqrt).reshape([conv.out_channels, 1, 1, 1])
        b = (b - mean) / var_sqrt * beta + gamma
        fused_conv = nn.Conv2d(conv.in_channels,
                               conv.out_channels,
                               conv.kernel_size,
                               conv.stride,
                               conv.padding,
                               bias=True)
        fused_conv.weight = nn.Parameter(w)
        fused_conv.bias = nn.Parameter(b)
        return fused_conv

    children = list(m.named_children())
    c = None
    cn = None

    for name, child in children:
        if isinstance(child, nn.BatchNorm2d):
            bc = fuse(c, child)
            m._modules[cn] = bc
            m._modules[name] = DummyModule()
            c = None
        elif isinstance(child, nn.Conv2d):
            c = child
            cn = name
        else:
            fuse_module(child)



# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import os
import torch.nn
import numpy as np
import matplotlib.pyplot as plt

from .quant_base import QuantBase


class QuantDebug(QuantBase):
    def __init__(self, name, module, bit_width, logger, log_path):
        super().__init__(name, module, bit_width, logger)
        self.log_path = log_path
        self.real_bins = 254
        self.real_weight_max = 0.0
        self.real_weight_min = 0.0
        self.real_activation_max = 0.0
        self.real_activation_min = 0.0
        self.real_activation_distribution = np.zeros(self.real_bins)
        self.real_activation_distribution_edge = None

    def init_boundary(self):
        # init activation boundary
        max_val = np.max(self.activation)
        min_val = np.min(self.activation)
        self.real_activation_max = max(max_val, self.real_activation_max)
        self.real_activation_min = min(min_val, self.real_activation_min)

    def init_histograms(self):
        if self.activation_positive:
            real_hist, real_hist_edge = np.histogram(
                self.activation,
                bins=self.real_bins,
                # use 2**(1 - self.bit_width) not zero,
                # because relu's activation have so many zero values, about 50~90%.
                range=(2**(1 - self.bit_width),
                       min(2**(self.bit_width - 1), self.real_activation_max)))
        else:
            real_hist, real_hist_edge = np.histogram(
                self.activation,
                bins=self.real_bins,
                range=(max(-2**(self.bit_width - 1), self.real_activation_min),
                       min(2**(self.bit_width - 1), self.real_activation_max)))
        self.real_activation_distribution_edge = real_hist_edge
        self.real_activation_distribution += real_hist

    def quantize_weight(self):
        try:
            # init real boundary
            # module can do not have weight, eg. relu
            weight_data = self.module.weight.cpu().data.flatten().numpy()
            self.real_weight_max = np.max(weight_data)
            self.real_weight_min = np.min(weight_data)
            # fig
            plt.title(self.name + " real weight histogram")
            plt.hist(weight_data, bins=100, density=True)
            plt.savefig(
                os.path.normpath(self.log_path + '/' + self.name +
                                 "_real_weight.png"))
            plt.clf()
        except:
            pass

    def quantize_activation(self):
        # fig
        self.real_activation_distribution /= sum(
            self.real_activation_distribution)
        plt.title(self.name + " real activation histogram")
        plt.bar(self.real_activation_distribution_edge[1:],
                self.real_activation_distribution)
        plt.savefig(
            os.path.normpath(self.log_path + '/' + self.name +
                             "_real_activation.png"))
        plt.clf()



# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import numpy as np

from .debug import QuantDebug
from .quant_base import *


class QuantInput(QuantDebug):
    def __init__(self, name, module, bit_width, logger, log_path, data_loader):
        super().__init__(name, module, bit_width, logger, log_path)
        self.weight_fl = -1
        self.data_loader = data_loader

    def debug_input(self):
        for index, (data, label) in enumerate(self.data_loader):
            self.activation = data.flatten().numpy()
            self.init_boundary()

        for index, (data, label) in enumerate(self.data_loader):
            self.activation = data.flatten().numpy()
            self.init_histograms()

        self.quantize_activation()

    def quantize_input(self):
        for index, (data, label) in enumerate(self.data_loader):
            data = data.flatten().numpy()
            hist, hist_edge = np.histogram(data,
                                           bins=self.bins_num,
                                           range=(-1 * self.activation_max,
                                                  self.activation_max))
            self.activation_distribution += hist
        self.activation_fl = kld_based_dynamics_double(
            self.activation_distribution, self.bit_width)


import os
import copy
import math
import numpy as np

from scipy import stats
from ..utils.quant_utils import float2fixed


class QuantBase:
    def __init__(self, name, module, bit_width, logger):
        self.name = name
        self.module = module
        self.bit_width = bit_width
        self.logger = logger

        self.weight_fl = 0
        self.activation_fl = 0

        self.activation = None
        self.activation_positive = True
        self.bins_num = 2**(2 * self.bit_width - 1)
        self.activation_max = 2**(bit_width - 1) - 1
        self.activation_distribution = np.zeros(self.bins_num)
        self.activation_distribution_edge = None

        self.hook_handle = None

    def eval_mode(self):
        if self.hook_handle is not None:
            self.hook_handle.remove()
        self.register_eval_hook()

    def register_get_activation_hook(self):
        def hook(module, input, output):
            self.activation = output.cpu().detach().data.flatten().numpy()

        self.module.register_forward_hook(hook)
        self.logger.info(self.name + " will be quantized.")

    def register_eval_hook(self):
        def hook(module, input, output):
            output = float2fixed(output, self.bit_width, self.activation_fl)
            return output

        if self.activation_fl is not -1:
            self.module.register_forward_hook(hook)

    def init_histograms(self):
        hist, hist_edge = np.histogram(self.activation,
                                       bins=self.bins_num,
                                       range=(-1 * self.activation_max,
                                              self.activation_max))
        self.activation_distribution += hist
        self.activation_distribution_edge = hist_edge

    def adaptive_kld(self):
        # -1, 1
        adaptive_zone = copy.deepcopy(
            self.activation_distribution[self.bins_num // 2 -
                                         2**(self.bit_width -
                                             1):self.bins_num // 2 +
                                         2**(self.bit_width - 1)])
        activation_sum = sum(self.activation_distribution)
        adaptive_zero = 2**(self.bit_width - 1)
        # handcraft rule
        alpha = 0
        beta = 2**(2 - self.bit_width) / 8
        weight = 0
        for i in range(self.bit_width - 1, -1, -1):
            weight = sum(adaptive_zone[adaptive_zero - 2**
                                       (self.bit_width - 1 - i):adaptive_zero +
                                       2**(self.bit_width - 1 - i)]) - weight
            # handcraft rule
            if weight / activation_sum > 0.1:
                beta *= 2
            # handcraft rule
            alpha += (2**(i + 1)) * (weight / activation_sum)

        def adaptive_func(alpha, beta, x):
            # 防止计算指数时溢出
            x /= 128 * 8
            # 防止计算指数时溢出
            if x > 0:
                param = min(20, max(-20, -1 * alpha * (-1 * x + beta)))
                return 1 - 1 / (1 + (math.exp(param)))
            else:
                param = min(20, max(-20, -1 * alpha * (x + beta)))
                return 1 - 1 / (1 + (math.exp(param)))

        adaptive_scale = [
            adaptive_func(alpha, beta, x - self.bins_num // 2)
            for x in range(self.bins_num)
        ]

        self.activation_distribution *= adaptive_scale

    def quantize_weight(self):
        weight_data = self.module.weight.cpu().data.flatten().numpy()
        # init kld
        hist, hist_edge = np.histogram(weight_data,
                                       bins=self.bins_num,
                                       range=(-1 * self.activation_max,
                                              self.activation_max))
        min_kld_fl = kld_based_dynamics_double(hist, self.bit_width)
        self.weight_fl = min_kld_fl
        self.module.weight.data = float2fixed(self.module.weight,
                                              self.bit_width, self.weight_fl)

    def quantize_activation(self):
        # pick fl which minimizes KL divergence
        if self.activation_positive:
            self.activation_fl = kld_based_dynamics_single(
                self.activation_distribution, self.bit_width)
        else:
            self.adaptive_kld()
            self.activation_fl = kld_based_dynamics_double(
                self.activation_distribution, self.bit_width)


def kld_based_dynamics_single(distribution, bit_width):
    kld = np.zeros(bit_width)
    # only calculate positive
    zero_bin = 2**(2 * bit_width - 2)
    distribution = distribution[zero_bin + 1:]

    for i in range(bit_width):
        range_bin = 2**(i + bit_width - 2)
        p = copy.deepcopy(distribution[:range_bin]).astype(np.float64)
        q = copy.deepcopy(distribution[:range_bin]).astype(np.float64)
        # p[0] += sum(distribution[:zero_bin - range_bin])
        p[-1] += sum(distribution[range_bin:])
        is_nonzeros = (p != 0).astype(np.int64)
        #extended q
        if i != 0:
            for j in range(0, 2**(i + bit_width - 1), 2**(i - 1)):
                step = 2**(i - 1)
                norm = sum(is_nonzeros[j:j + step])
                q[j:j + step] = sum(q[j:j + step]) / norm if norm != 0 else 0
        q[p == 0] = 0
        p[p == 0] = 0.000001
        q[q == 0] = 0.000001
        # calculate kl_divergence between q and p
        kld[i] = stats.entropy(p, q)

    min_kld_fl = np.argmin(kld)

    return min_kld_fl


def kld_based_dynamics_double(distribution, bit_width):
    kld = np.zeros(bit_width)
    zero_bin = 2**(2 * bit_width - 2)
    # distribution[16300:16500] = 0

    for i in range(bit_width):
        range_bin = 2**(i + bit_width - 1)
        p = copy.deepcopy(distribution[zero_bin - range_bin:zero_bin +
                                       range_bin]).astype(np.float64)
        q = copy.deepcopy(distribution[zero_bin - range_bin:zero_bin +
                                       range_bin]).astype(np.float64)
        p[0] += sum(distribution[:zero_bin - range_bin])
        p[-1] += sum(distribution[zero_bin + range_bin:])
        is_nonzeros = (p != 0).astype(np.int64)
        #extended q
        if i != 0:
            for j in range(0, 2**(i + bit_width), 2**i):
                if j == 0 or j == 2**(i + bit_width) - 2**i:
                    step = 2**(i - 1)
                else:
                    step = 2**i
                norm = sum(is_nonzeros[j:j + step])
                q[j:j + step] = sum(q[j:j + step]) / norm if norm != 0 else 0
        q[p == 0] = 0
        p[p == 0] = 0.000001
        q[q == 0] = 0.000001
        # calculate kl_divergence between q and p
        kld[i] = stats.entropy(p, q)

    min_kld_fl = np.argmin(kld)

    return min_kld_fl



# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch.nn
from ..quant_base import QuantBase


class QuantBatchNorm(QuantBase):
    def __init__(self, name, module, bit_width, logger):
        super().__init__(name, module, bit_width, logger)

    def quantize_weight(self):
        pass

    def quantize_activation(self):
        pass


class QuantLayerNorm(QuantBase):
    def __init__(self, name, module, bit_width, is_debug):
        super().__init__(name, module, bit_width, is_debug)



# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch.nn as nn


class TorchAdd(nn.Module):
    def __init__(self):
        super(TorchAdd, self).__init__()

    def forward(self, x, y):
        return x + y


class TorchMinus(nn.Module):
    def __init__(self):
        super(TorchMinus, self).__init__()

    def forward(self, x, y):
        return x - y


class TorchMatMul(nn.Module):
    def __init__(self):
        super(TorchMatMul, self).__init__()

    def forward(self, x, y):
        return torch.matmul(x, y)


class TorchDotMul(nn.Module):
    def __init__(self):
        super(TorchDotMul, self).__init__()

    def forward(self, x, y):
        return x * y


class TorchDiv(nn.Module):
    def __init__(self):
        super(TorchDiv, self).__init__()

    def forward(self, x, y):
        return x / y


class TorchSin(nn.Module):
    def __init__(self):
        super(TorchSin, self).__init__()

    def forward(self, x):
        return torch.sin(x)


class TorchCos(nn.Module):
    def __init__(self):
        super(TorchCos, self).__init__()

    def forward(self, x):
        return torch.cos(x)


class TorchSoftmax(nn.Module):
    def __init__(self):
        super(TorchSoftmax, self).__init__()

    def forward(self, x):
        return torch.softmax(x)


# global var
TorchFuncTuple = (TorchAdd, TorchMinus, TorchMatMul, TorchDotMul, TorchDiv,
                  TorchSin, TorchCos, TorchSoftmax)